import java.io.*;
import java.util.*;


/**
 * Access verifier for hosts. It checks whether access from/to a host 
 * should be accepted or denied, by comparing with the tables in memory.
 */ 
public class AccessVerifier {

    final static int ALLOW     = 0;
    final static int DENY      = 1;
    final static int NOT_FOUND = 2;

    final static int NO_META = 0;
    final static int BEGINS_META = 1;
    final static int ENDS_META = 2;
    final static int BEGINS_ENDS_META = 3;


    /** A table of integer constants (ALLOW or DENY), specifying the access */ 
    private ArrayList accesses;
    /** A table of Strings, specifying the host */
    private ArrayList hosts;
    /** Builds a new AccessVerifier following the directives written in the
     * ASCII file passed as parameter.
     * @param filename  the name of the file containing the directives
     */ 
    public AccessVerifier(String filename) 
	                    throws IOException, UnrecognizedCfglineException {
	accesses = new ArrayList();
	hosts = new ArrayList();
	LineNumberReader reader = 
	    new LineNumberReader(new FileReader(filename));

	// Reads the file and stocks the access/host pairs in the tables
	String line = new String();
	while ((line = reader.readLine()) != null) {
	    StringTokenizer st = new StringTokenizer(line);
	    while (st.hasMoreTokens()) {
		String token1 = st.nextToken();
		if (token1.equals("allow"))
		    accesses.add(new Integer(ALLOW));
		else if (token1.equals("deny"))
		    accesses.add(new Integer(DENY));
		else
		    throw new UnrecognizedCfglineException(
				     token1, reader.getLineNumber(), filename);

		String token3 = st.nextToken();
		hosts.add(token3);
	    }
	}

    }


    /**
     * Checks the access directive for a given host.
     * @param hostname  the host's name
     * @return  a constant (ALLOW, DENY or NOT_FOUND) describing the access,
     * following the directives of this AccessVerifier object 
     */
    private int getAccess(String hostname) throws NoSuchElementException {
	int index = 0;

	while (true) {
	    String hostEntry = new String((String) hosts.get(index));

	    if (hostEntry.indexOf('?') != -1) {
		// host entry in the table contains metacharacter '?'
		if (hostEntry.length() == hostname.length()) {
		    byte[] hn = new byte[hostname.length()];
		    byte[] hE = new byte[hostEntry.length()];
		    hn = hostname.getBytes();
		    hE = hostEntry.getBytes();
		    boolean match = true;
		    for (int i = 0; i < hostname.length(); i++) 
			if (hE[i] != '?')
			    if (hn[i] != hE[i]) { 
				match = false; 
				break; 
			    }
		    if (match == true) break;
		}
	    }
	    else {
		int modeMatch = NO_META;
		if (hostEntry.startsWith("*")) { 
		    // host entry in the table begins with metacharacter '*' 
		    hostEntry = hostEntry.substring(1);
		    modeMatch += BEGINS_META; 
		}
		if (hostEntry.endsWith("*")) { 
		    // host entry in the table ends with metacharacter '*' 
		    hostEntry = hostEntry.substring(0, hostEntry.length()-1);
		    modeMatch += ENDS_META;
		}

		if (modeMatch == NO_META)
		    if (hostname.equals(hostEntry)) break;
		if (modeMatch == BEGINS_ENDS_META)
		    if (hostname.indexOf(hostEntry) != -1) break;
		if (modeMatch == BEGINS_META) 
		    if (hostname.endsWith(hostEntry)) break;
		if (modeMatch == ENDS_META) 
		    if (hostname.startsWith(hostEntry)) break;
	    }

	    if (index == hosts.size()-1) return NOT_FOUND;
	    index++;
	}

	// Hostname matches with an entry in the hosts table, 
	// so the result is returned 

	return ((Integer) accesses.get(index)).intValue();
    }
        
    /**
     * Checks whether a client is acceptable or not.
     * @param clientName  the client's name
     * @return  true if the client is acceptable, false elsewhere
     */ 
     public boolean shouldAcceptThisClient(String clientName) {
	switch(getAccess(clientName)) {
	case ALLOW: return true;
	case DENY: return false;
	case NOT_FOUND: return false;
	}
	return false;    // for compiling only (should never go there)
    }


    /**
     * Checks whether a server is acceptable or not.
     * @param serverName  the server's name
     * @return  true if the server is acceptable, false elsewhere
     */ 
    public boolean shouldAcceptThisServer(String serverName) {
	switch(getAccess(serverName)) {
	case ALLOW: return true;   
	case DENY: return false;
	case NOT_FOUND: return true;
	}
	return true;     // for compiling only (should never go there)
     }


    /**
     * Sends a message which informs that the proxy 
     * refuses connection from client.
     * @param is  the stream of the connection
     */ 
    public static void warnDeniedClient(OutputStream is) {
	String warnMsg = "<HTML>\n<HEAD>\n" +
	    "<TITLE>Connection Refused</TITLE>\n" +
	    "</HEAD>\n<BODY>\n" +
	    "<H1>Connection Refused</H1>\n<BR>" +
	    "Sorry, you don't have the right to connect to this proxy." +
	    "\n<BR><HR>\n" +
	    "<I>Generated by RAMM Proxy at " + new Date() + "</I>\n" +
	    "</BODY></HTML>\n";
	writeWarningMsg(is, warnMsg);
    }


    /**
     * Sends a message which informs that the proxy 
     * denies access to a server.
     * @param is  the stream of the connection
     * @param deniedSite  the name of the denied server 
     */
    public static void warnDeniedServer(OutputStream is, String deniedSite) {
	String warnMsg = "<HTML>\n<HEAD>\n" + 
	    "<TITLE>Access Denied</TITLE>\n" +
	    "</HEAD>\n<BODY>\n" +
	    "<H1>Access Denied</H1>\n<BR>" +
	    "Sorry, you cannot access " + deniedSite + 
	    " via this proxy.\n<BR><HR>\n" +
	    "<I>Generated by RAMM Proxy at " + new Date() + "</I>\n" +
	    "</BODY></HTML>\n";
	writeWarningMsg(is, warnMsg);
    }

    /**
     * Sends an HTML message into the connection stream.
     * @param is  the stream of the connection
     * @param warnMsg  the body message
     */
    private static void writeWarningMsg(OutputStream is, String warnMsg) {
	String warnHeader = "HTTP/1.1 200 OK\n" +
	    new Date() + "\n" +
	    "Server: RAMM Proxy v1.0\n" +
	    "Accept-Ranges: bytes\n" +
	    "Content-Length: " + warnMsg.length() + "\n" +
	    "Content-Type: text/html\n" +
	    "\n\r";
	try {
	    is.write((warnHeader + warnMsg).getBytes());
	} catch (IOException e) { System.out.println(e); }
    }


}
