import java.net.*;
import java.io.*;
import java.util.*;


/**
 * Stream reader for HTTP messages (requests and responses).
 */
public abstract class HTTPMessageReader {

    /** The header of the message (blank line included) */
    protected ArrayList header;

    /** The body of the message */
    protected ArrayList body;

    /** The whole message obtained from the stream */
    protected byte[] message;

    /** Start/Status line: The version of HTTP protocol, eg. HTTP/1.1 */
    String versionProtocol;

    /** Header field: The length of message body */
    protected int contentLength = -1;

    /** Position of body in the message */
    protected int offset;

    /** Header field: Connection */
    protected String connection;

    /** Flag that informs whether the connection must be closed or not */
    protected boolean mustCloseConnection = true;


    /**
     * Reads inside the stream until the "\r\n" couple that
     * separes header from body message.
     * @param distream  the stream of connection
     */
    protected void retrieveHeader(DataInputStream distream) 
	                                                   throws IOException {
	header = new ArrayList();
	byte data;
	offset = 0;
	while (true) {
	    data = (byte) distream.read();
	    if (data == '\r') {
		header.add(new Byte(data));
		data = (byte) distream.read();
		if (data == '\n') {
		    header.add(new Byte(data));
		    data = (byte) distream.read();
		    if (data == '\r') {
			header.add(new Byte(data));
			data = (byte) distream.read();
			if (data == '\n') {
			    header.add(new Byte(data));
			    break;
			}
		    }
		}
	    }
	    header.add(new Byte(data));
	}

	if (ProxyMain.TRACEON_MSGS)
	    System.out.print(new String(toArray(header)));
    }
	

    
    /**
     * Receives the body of the message (if exists) from the stream.
     * @param distream  the stream of connection 
     */
    protected void retrieveBody(DataInputStream distream) throws IOException {

	/* HTTP/1.0 compliant: sets a buffer enough large for the data.
	   HTTP/1.0 marks the end of a message by closing the connection
	   (no keep-alive), thus does not need a Content-Length header
	   which is obligatory in HTTP/1.1
	*/ 
	if (contentLength == -1) contentLength = 1024 * 1024;

	byte[] bbody = new byte[contentLength];
	distream.readFully(bbody, 0, contentLength);
	body = new ArrayList();
	for (int i = 0; i < contentLength; i++) body.add(new Byte(bbody[i]));

	if (ProxyMain.TRACEON_FULLMSGS)
	    System.out.println(new String(toArray(body)) + "\n");
	else if (ProxyMain.TRACEON_MSGS)
	    System.out.println("[" + body.size() + " bytes of data body]\n");
	
	concatenateHeaderBody();
    }


    /**
     * Concatenates the header and the body in a whole message.
     */
    private void concatenateHeaderBody() {
	message = new byte[header.size() + body.size()];
	int c = 0;
	for (int c1 = 0; c1 < header.size(); c1++)
	    message[c++] = ((Byte) header.get(c1)).byteValue();
	for (int c2 = 0; c2 < body.size(); c2++)
	    message[c++] = ((Byte) body.get(c2)).byteValue();
    }


    /**
     * Converts an ArrayList of Bytes into an array of bytes.
     * @param arraylist  an ArrayList
     * @return  an array of bytes
     */ 
    public static byte[] toArray(ArrayList arraylist) {
	byte[] array = new byte[arraylist.size()];
	for (int i = 0; i < arraylist.size(); i++) 
	    array[i] = ((Byte) arraylist.get(i)).byteValue();
	return array;
    }

    /**
     * Erases from the header the line containing a command.
     * @param cmd the command that identifies the line
     * @param nbrParam the number of parameter to erase after the command cmd
     */
    public void eraseCommand(String cmd, int nbrParam) {
	String string; 
	byte[] bytes;
	ArrayList sal = new ArrayList();
	StringTokenizer st = 
	    new StringTokenizer(new String(toArray(header)), "\n", true);
	while (st.hasMoreTokens()) {
	    string = new String(st.nextToken());
	    if (string.equals(cmd)) {
		for (int i=0; i<nbrParam; i++) st.nextToken();
	    }
	    else {
		bytes = string.getBytes();
		for (int i = 0; i < bytes.length; i++) 
		    sal.add(new Byte(bytes[i]));
	    }
	}
	header = sal;
	
	concatenateHeaderBody();
    } 
}
