import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Represents a connection between a client and a server; a static thread 
 * manager manages the number of active and existing threads, and a static 
 * ServerSocket waits for connections.  
 */
public class ProxyConnection implements Runnable {
    /** manages the number of active and existing threads */
    static ThreadManager man;
    /** all the connection are catched on this ServerSocket */
    static ServerSocket serverSocket;
    /** the interface with the server */
    Socket serverSideSocket; 
    /** the interface with the client */
    Socket clientSideSocket;
    /** inputstream from the client */
    InputStream clientSideIn;
    /** outputstream to the client */    
    OutputStream clientSideOut;
    /** inputstream from the server */
    InputStream serverSideIn;
    /** outputstream to the server */
    OutputStream serverSideOut;
    /** to verify if the connection must be closed */
    boolean mustCloseConnection;


    /**
     * Default constructor, used by the ThreadManager.
     */
    ProxyConnection() {}


    /**
     * Starts the proxy on the port passed as parameter,
     * or, if already used,  on a free port.
     * @param port the TCP port where waiting for connections 
     */ 
    ProxyConnection(int port) throws IOException {
	try {
	    man = new ThreadManager();
	    serverSocket = new ServerSocket(port);
	}
	catch (IOException e) {
	    serverSocket = new ServerSocket(0);
	}
	ProxyMain.PROXY_HOST = 
	    this.serverSocket.getInetAddress().getHostName();
	System.out.println("Proxy running on " + 
			   this.serverSocket.getInetAddress().getHostName() +
			   ", port " +
			   this.serverSocket.getLocalPort());
    }


    /**
     * Changes the bounds for the number of threads.
     * It asks the thread manager to do it.
     * @param maxThreads maximum number of existing Threads
     * @param minThreads minimum number of existing Threads
     * @param maxSpareThreads maximum number of disactivated Threads
     * @param minSpareThreads minimum number of disactivated Threads
     * @exception IllegalValuesException
     */ 
    void setAttributs(int maxThreads, int minThreads, 
		      int maxSpareThreads, int minSpareThreads) 
                                                throws IllegalValuesException {
	man.setAttributs(maxThreads, minThreads, 
			 maxSpareThreads, minSpareThreads);
	// System.out.println(man.printStatus());
    }


    /**
     * Asks the thread manager to create minThreads threads.
     * @exception IOException, InstantiationException, IllegalValuesException
     */
    void launch() throws IOException, InstantiationException, 
	IllegalAccessException {
	man.initialize(ProxyConnection.class);
    }

    /**
     * Waits for a connection and manages it until either the client 
     * or the server wants close it.
     */
    public void run() {
	String client;   
	String server;   
	HTTPResponseReader httpResponse;
	HTTPRequestReader httpRequest;
	while (!man.mustDestroyThread()) {

	    mustCloseConnection = true;
	    try {
		// waits for a connection...
		clientSideSocket = serverSocket.accept();
		// ...connection arrived!
		man.oneMoreActiveThread();

		// filtering: client's permissions verification 
		client = clientSideSocket.getInetAddress().getHostName();
		if (!ProxyMain.accessClient.shouldAcceptThisClient(client)) {
		    if (ProxyMain.TRACEON_FILTERS)
			System.out.println("Client " + client + 
					   " is not welcome on this proxy");
		    AccessVerifier.warnDeniedClient(
					   clientSideSocket.getOutputStream());
		    // clientSideSocket.close();
		    // commented, as it will result in a "broken pipe" error
		    continue;
		}	
		// asks the ThreadManager (man) to manage the number of threads
		if (man.mustCreateThread()) {
		    man.newThread(ProxyConnection.class);
		}

		// Reads the HTTP message on the clientSideIn stream		
		clientSideIn = clientSideSocket.getInputStream();
		
		httpRequest = new HTTPRequestReader(clientSideIn);
		mustCloseConnection = httpRequest.mustCloseConnection;
		
		// filtering: server's permissions verification
		server = httpRequest.host;
		if (!ProxyMain.accessServer.shouldAcceptThisServer(server)) {
		    if (ProxyMain.TRACEON_FILTERS)
			System.out.println("Client " + client + 
					   " cannot access server " + server);
		    // logging...
		    ProxyMain.log.writeLine("[" +
					    InetAddress.getByName(client).
					    getHostAddress() + 
					    "] [" + new Date() +
					    "] [" + httpRequest.startLine +
					    "] [DENIED]");
		    AccessVerifier.warnDeniedServer(
				   clientSideSocket.getOutputStream(), server);
		    // clientSideIn.close();
		    // clientSideSocket.close();
		    continue;
		}

		if (ProxyMain.auth.authRequired(server)) { 
		    //if authorization required, see the auth
		    if (httpRequest.auth == null) {
			AuthorizationVerifier.warnAuthRequired(
			       clientSideSocket.getOutputStream(), "defendu");
			continue;
		    }
		    else if (!ProxyMain.auth.passwdVerify(
					         server, httpRequest.auth)) {
			AuthorizationVerifier.warnForbidden(
					  clientSideSocket.getOutputStream());
			continue;
		    }
		    else ; // password OK, process the request 
		}
		    		
		// logging...
		ProxyMain.log.writeLine("[" +
					InetAddress.getByName(client).
					getHostAddress() + 
					"] [" + new Date() +
					"] [" + httpRequest.startLine +
					"] [ACCEPTED]");
		

		// Sends the message to the server
		serverSideSocket = new Socket(InetAddress.getByName(
						         httpRequest.host),80);
		// a useful time-out
		//serverSideSocket.setSoTimeout(ProxyMain.TIMEOUT); 

		serverSideOut = serverSideSocket.getOutputStream();
		serverSideOut.write(httpRequest.message);

		serverSideIn = serverSideSocket.getInputStream();
		clientSideOut = clientSideSocket.getOutputStream();
		// Reads the server-side message and forwards it to the client
		do { 
		    httpResponse = new HTTPResponseReader(serverSideIn);
		    mustCloseConnection = httpResponse.mustCloseConnection;
		    clientSideOut.write(httpResponse.message);
		} while (httpResponse.statusCode.startsWith("1")); 
		// control message, wait for an other one
		

		// First request-response managed. 
		// Now, if !mustCloseConnection, KEEP-ALIVE
		while (!mustCloseConnection) {
		    // reads another message...
		    httpRequest = new HTTPRequestReader(clientSideIn);
		    mustCloseConnection = httpRequest.mustCloseConnection;
		    // and manages  another response...
		    do { 
		    httpResponse = new HTTPResponseReader(serverSideIn);
		    mustCloseConnection = httpResponse.mustCloseConnection;
		    clientSideOut.write(httpResponse.message);
		    } while (httpResponse.statusCode.startsWith("1"));
		}

	    }
	    catch (UnknownHostException uh) {System.out.println("Unknown host");}
	    catch (InterruptedIOException iio) {System.out.println("Timeout - connection closed");}
	    catch (IOException io) {}
	    catch (InstantiationException inst) {}
	    catch (IllegalAccessException illacc) {}
	    finally { // if there is a problem, CLOSE the connection!! 
		try {
		    clientSideSocket.close();
		    clientSideIn.close();
		    serverSideSocket.close();
		    serverSideOut.close();
		    serverSideIn.close();
		    clientSideOut.close();
		}
		catch (NullPointerException npe) {}
		catch (IOException ioe) {}
		man.oneLessActiveThread();
		// man.printStatus();
	    }
	}
	man.oneLessExistingThread();
	// man.printStatus();
    }
    

}

