/**
 * Manager for the threads. Every threads is used for a connection
 * for downloading a part of the ressource.
 */
class ThreadManager {
    private int maxThreads = 50;       
    private int minThreads = 7;       
    private int maxSpareThreads = 5;  
    private int minSpareThreads = 5;  
    static int existingThreads = 0;  
    static int activeThreads = 0;    


    /**
     * Creates a new ThreadManager with default values as attributs.
     * maxThreads = 50;    
     * minThreads = 5;       
     * maxSpareThreads = 5;  
     * minSpareThreads = 5;    
     * It doesn't create new Threads.
     * Threads must be created using the initialize() method.
     */
    public ThreadManager() {}


    /**
     * It sets the bounds for the Threads, but it doesn't create new Threads.
     * Threads must be created using the initialize() method.
     * @param maxThreads maximum number of existing Threads
     * @param minThreads minimum number of existing Threads
     * @param maxSpareThreads maximum number of disactivated Threads
     * @param minSpareThreads minimum number of disactivated Threads
     */
    public ThreadManager(int maxThreads, int minThreads, 
			 int maxSpareThreads, int minSpareThreads)
	throws IllegalValuesException {
	this.setAttributs(maxThreads, minThreads, 
			  maxSpareThreads, minSpareThreads);
    }


   /**
     * Method that creates minThreads Threads
     * @param c Class extending Runnable, every new Thread manages a 
     * new object of this class
     */
    public void initialize(Class runnableClass) throws InstantiationException, 
	                                              IllegalAccessException {
	for (int i=0; i<minThreads; i++) 
	    newThread(runnableClass);
    }


    /**
     * This method changes the bounds for the number of Threads
     * @param maxThreads new maximum number of existing Threads
     * @param minThreads new minimum number of existing Threads
     * @param maxSpareThreads new maximum number of disactivated Threads
     * @param minSpareThreads new minimum number of disactivated Threads
     * @exception IllegalValuesException
     */
    public void setAttributs(int maxThreads, int minThreads, int maxSpareThreads, int minSpareThreads)
	throws IllegalValuesException {
	// If values are not correct, an IllegalValuesException is thrown 
	if (minThreads > maxThreads) 
	    throw new IllegalValuesException(
				     "Trying to set minThreads > maxThreads");
	if (minThreads < minSpareThreads) 
	    throw new IllegalValuesException(
				"Trying to set minThreads < minSpareThreads"); 
	if (minSpareThreads > maxSpareThreads) 
	    throw new IllegalValuesException(
			   "Trying to set minSpareThreads > maxSpareThreads");
	if (minSpareThreads < 0) 
	    throw new IllegalValuesException(
					  "Trying to set minSpareThread < 0");
	// ...else attributs are updated
	this.maxThreads=maxThreads;
	this.minThreads=minThreads;
	this.maxSpareThreads=maxSpareThreads;
	this.minSpareThreads=minSpareThreads;
    }


    /**
     * This method returns a String describing the number of existing 
     * and activated Threads.  
     * @return a String with number and status of threads
     */
    public synchronized String printStatus() {
        return "Threads: " + existingThreads + ", Actives: " + activeThreads;
    }

   
    /**
     * Method that creates a new Thread, starts it and update attributs 
     * of the ThreadManager. 
     * @param objectRunnable An object instance of Runnable, 
     * managed by a new Thread
     * @exception InstantiationException, IllegalAccessException
     */
    public synchronized void newThread(Class runnableClass) throws InstantiationException, IllegalAccessException {
        Thread t=new Thread((Runnable) runnableClass.newInstance());
	t.setDaemon(true);
	existingThreads++;
	t.start();
    }


    /**
     * This method increments the number of activated Threads
     * Attention!! It doesn't activate a Thread.
     */
    public void oneMoreActiveThread() {
	activeThreads++;
    }


    /**
     * This method decrements the number of activated Threads
     * Attention!! It doesn't deactivate a Thread.
     */
    public void oneLessActiveThread() {
	activeThreads--;
    }


    /**
     * This method decrements the number of existing Threads
     * Attention!! It doesn't destroy a Thread.
     */
    public void oneLessExistingThread() {
	existingThreads--;
    }
   

    /**
     * This method return true if a new Thread must be created
     */
    public synchronized boolean mustCreateThread() {
	return (existingThreads<minThreads)
	    ||(existingThreads<maxThreads && activeThreads>(existingThreads-minSpareThreads));
    }


    /**
     * This method return true if an existing Thread must be destroyed
     */
    public synchronized boolean mustDestroyThread() {
	return (existingThreads>maxThreads)
	    ||(existingThreads>minThreads && activeThreads<(existingThreads-maxSpareThreads));
    }

}

